package vt.cs.db.filedb;

import java.net.*;
import java.util.*;
import java.io.*;
import vt.cs.db.*;

           
/**
 * A DiskFileTable gets constructed with the name (URL) of the "codebook"
 * file (which is in SGML) for a single data file.  This file is parsed
 * to get the following information. <UL>
 * <LI> The "name" of this data
 * <LI> Name of the data file (an ASCII file with "rows" separated by newlines)
 * <LI> Whether the rows represent variables or records
 * <LI> The character(s) used as delimiters
 * <LI> The string used for missing values
 * <LI> Attribute-value pairs for this data
 * <LI> A list of variables with information about each, 
 *      in order of how the variables appear in the data file
 * </UL>
 * The information about each variable includes: <UL>
 * <LI> Its short name (required)
 * <LI> Type that should be used to store values (required).
 * <LI> Its description--a longer string (optional)
 * <LI> Value label descriptions (for categorical variables)
 * </UL>
 *
 */
class DiskFileTable extends Table {
    private URL codebookFile = null;
    private URL dataFile = null;     
    private DataInputStream datastream = null;

    private Hashtable varIndices = null; // Index to variable indices by name
    private Vector variables = null;     // Array of "Variable" objects with 
                                         // all info. about the variables.

    private int numVars = 0;    
    private int numRecords = 0;
    private boolean rowsAreVars;
    private String dataDelim = null;
    private String missingVal = null;
    private Vector lineOffsets = null;

    private StringTokenizer linetok = null; // For anyone to use

    public DiskFileTable(String url) throws DatabaseException { 
        setName(url);       // For now, just use filename
        varIndices = new Hashtable();
        variables = new Vector();
        try {
            // Open codebook file & parse
            codebookFile = new URL(url);
            numVars = parseCodebookFile();
        }
        catch (Exception e) {
            e.printStackTrace();   // Change this--shouldn't have two levels
            throw new DatabaseException
                ("Error reading from codebook file " + codebookFile);
        }
        try {
            // Now open dataFile, calculate offset positions, 
            // count number of records, and do a test
            lineOffsets = new Vector();
            numRecords = parseDataFile();
            get(1,1);  // Make sure there's data & not just an error message
        } catch(Exception e) { 
           e.printStackTrace();
           throw new DatabaseException
               ("Error reading from data file " + dataFile);
        }
    }

    public long recordCount() { return numRecords; }
        
    public String[] variableNames() {
        // return varIndices.keys();  // Returns variable names in weird order
        String[] temp = new String[numVars];
        for (int i = 0; i < numVars; i++) temp[i] = getVariable(i).name;
        return temp;
    }

    public long rowcount() { return recordCount(); }
    public Enumeration columnNames(String type) { return null; }
    public Enumeration columnNames() { return null; }
    
    public String[] variableNames(String type) {
        String[] temp = new String[numVars];
                int count = 0;
        for (int i = 0; i < numVars; i++) {
            if (variableType(i).equalsIgnoreCase(type))
                temp[count++] = getVariable(i).name;
        }
        // Copy into array of proper size
        String[] retArray = new String[count];
        System.arraycopy(temp, 0, retArray, 0, count);

        return retArray;
    }

    public String description(String varName) throws DatabaseException {
        String desc = (getVariable(varIndex(varName))).description;
        if (desc != null) return desc;
        else return "No description available";
    }

    public Object get(int recNumber, int varNumber) throws DatabaseException {
        // Check against numRecords and numVars
        if (recNumber > numRecords)
            throw new DatabaseException("No such record.");
        if (varNumber > numVars)
            throw new DatabaseException("No such variable.");

        // Figure out which way we're facing
        int rowNum = rowsAreVars ? varNumber : recNumber;
        int elemNum = rowsAreVars ? recNumber : varNumber;

        String returnString = null;
        try {
            openDataFile();
            // seek in file to lineOffsets.elementAt(rowNum-1)
            skipToOffset(datastream, 
                ((Long)lineOffsets.elementAt(rowNum-1)).longValue());

            linetok = new StringTokenizer(datastream.readLine(), dataDelim);

            // then return the elemNum'th item on this line
            for (int i = 1; i++ < elemNum; linetok.nextToken());
            returnString = linetok.nextToken();
        }
        catch(Exception e) { e.printStackTrace(); }
        finally { closeDataFile(); }

        if (returnString.equals(missingVal)) return new MissingVal();
        else return classToType(returnString, getVariable(varNumber-1).type);
    }

    public Object get(int recNumber, String varName) throws DatabaseException {
        // Gets "row" number rowNumber of column "varName"
        return get(recNumber, varIndex(varName)+1);
    }

    public Object[] get(String varName) throws DatabaseException { 
        int varNum = varIndex(varName) + 1;
        if (varNum > numVars)
            throw new DatabaseException("No such variable.");
        if (rowsAreVars) return getRow(varNum, storedType(varNum-1));
        else return getColumn(varNum, storedType(varNum-1));
    }

    public Object[] get(int recNumber) throws DatabaseException { 
        if (recNumber > numRecords)
            throw new DatabaseException("No such record.");
        if (rowsAreVars) return getColumn(recNumber, null);
        else return getRow(recNumber, null);
    }

    private Object[] getRow(int rowNumber, Class elemType) {
        Object[] temp;
        if (rowsAreVars) temp = new Object[numRecords];
        else temp = new Object[numVars];
        try {
            openDataFile();
            // seek in file to lineOffsets.elementAt(rowNumber-1)
            skipToOffset(datastream,
                      ((Long)lineOffsets.elementAt(rowNumber-1)).longValue());
            linetok = new StringTokenizer(datastream.readLine(), dataDelim);

            // Now return the whole line, with the correct type
            for (int elemNum = 0; elemNum < temp.length; elemNum++) {
                String elemString = linetok.nextToken();
                if (elemString.equals(missingVal)) 
                    temp[elemNum] = new MissingVal();
                else if (rowsAreVars)
                    temp[elemNum] = classToType(elemString, elemType);
                else // Have to get the type
                    temp[elemNum] = classToType(elemString, storedType(elemNum));
            }
        }
        catch(Exception e) { e.printStackTrace(); }
        finally { closeDataFile(); }
        return temp;
    }

    private Object[] getColumn(int colNumber, Class elemType) {
        Object[] temp;
        if (rowsAreVars) temp = new Object[numVars];
        else temp = new Object[numRecords];
        try {
            openDataFile();
            // Go through every line of file, getting the colNumber'th element
            for (int i = 0; i < temp.length; i++) {
                linetok = new StringTokenizer(datastream.readLine(), dataDelim);
                for (int j = 1; j++ < colNumber; linetok.nextToken());
                String elemString = linetok.nextToken();
                // Now figure out its type
                if (elemString.equals(missingVal)) 
                    temp[i] = new MissingVal();
                else if (! rowsAreVars)
                    temp[i] = classToType(elemString, elemType);
                else // Have to get the type
                    temp[i] = classToType(elemString, storedType(i));
            }
        }
        catch(Exception e) { e.printStackTrace(); }
        finally { closeDataFile(); }

        return temp;
    }

    public void appendRow(Object[] rowData) throws DatabaseException {
        throw new DatabaseException("Method not supported.");
    }

    public String variableType(String varName) throws DatabaseException {
        return variableType(varIndex(varName));
    }

    public String variableAttribute(String varName, String att) 
        throws DatabaseException {
        if (attributes == null) return null;
        else {
            Variable currVar = getVariable(varIndex(varName));
            return (String)currVar.attributes.get(att.toUpperCase());
        }
    }
    
    private String variableType(int varIndex) {
        return storedType(varIndex).getName();
    }

    private Class storedType(int varIndex) {
        return getVariable(varIndex).type;
    }

    private int parseCodebookFile() throws IOException, MalformedURLException,
        DatabaseException {

        LineNumberInputStream cbStream = 
            new LineNumberInputStream(new BufferedInputStream(codebookFile.openStream()));
        // Put "name" from DATABASE tag 
        Tag currTag = getNextTag(cbStream);
        if (! currTag.name.equals("DATABASE"))
            throw new CodebookSyntaxException(cbStream.getLineNumber());
        // setName(currTag.attribute("NAME"));      // For right now, leave as filename

        // Get dataFile name from DATFILE tag
        currTag = getNextTag(cbStream);
        if (! currTag.name.equals("DATAFILE"))
            throw new CodebookSyntaxException(cbStream.getLineNumber());
        dataFile = new URL(codebookFile, currTag.attribute("NAME"));

        // Set rowsAreVars, dataDelim, and missingVal from DATAFILE tag
        rowsAreVars = (currTag.attribute("ROW").equalsIgnoreCase("RECORD"))
            ? false : true;     
        dataDelim = currTag.attribute("DELIM");
        missingVal = currTag.attribute("MISSING");

        // Put attributes from ATTR tags into "attributes" Hashtable
        for (currTag = getNextTag(cbStream); currTag.name.equals("ATTR"); 
            currTag = getNextTag(cbStream)) {
            setAttribute(currTag.attribute("NAME"), currTag.attribute("VAL"));
        }
    
        // Now read variable information from remaining tags
        int varCount = 0;
        Variable currVar = null;
        while(! currTag.name.equals("EOF")) {    // Loop until EOF
            if (currTag.name.equals("VARIABLE")) {   // VARIABLE tag
                Class varType = pickType(currTag.attribute("TYPE"));
                currVar = new Variable(currTag.attribute("NAME"), varType);
                variables.addElement(currVar);
                varIndices.put(currVar.name, new Integer(varCount++));
            }

            else if (currTag.name.equals("LABEL")) {   // LABEL tag
                if (currVar == null) 
                    throw new CodebookSyntaxException(cbStream.getLineNumber());
                currVar.description = currTag.attribute("STRING");
            }

            else if (currTag.name.equals("VALLABEL")) {  // VALLABEL tag
                if (currVar == null) 
                    throw new CodebookSyntaxException(cbStream.getLineNumber());
                int code = (new Integer(currTag.attribute("CODE"))).intValue();
                currVar.addValLabel(code, currTag.attribute("MEANING"));
            }

            else if (currTag.name.equals("ATTR")) {    // ATTR tag
                if (currVar == null)
                    throw new CodebookSyntaxException(cbStream.getLineNumber());
                String attrName = (String)currTag.attribute("NAME");
                currVar.addAttr(attrName.toUpperCase(), currTag.attribute("VALUE"));
            }
            // Ignore any other tags
            currTag = getNextTag(cbStream);
        }

        // Close codebook file
        cbStream = null;
        return varCount;
    }

    private int parseDataFile() throws IOException {
        int recordCount = 0;
        BufferedInputStream buffstream = 
            new BufferedInputStream(dataFile.openStream());

        // Scan through and find the offsets for each new line
        long pos = 0;       // Offset in file
        lineOffsets.removeAllElements();
        lineOffsets.addElement(new Long(pos));      // For first position
        boolean EOLRead = true;
loop:   while (true) {
            pos++;
            switch (buffstream.read()) {
             case -1:
                if (EOLRead) 
                // Take out last entry if file ended with newline
                lineOffsets.removeElement(lineOffsets.lastElement());
                break loop;
             case '\r':
                buffstream.mark(1);
                if (buffstream.read() != '\n') buffstream.reset();
                else pos++;
                // no break here on purpose
             case '\n': 
                lineOffsets.addElement(new Long(pos));
                EOLRead = true;
                break;
             default:
                EOLRead = false;
            }
        }

        buffstream = null;

        // Now count the number of records
        if (rowsAreVars) {    // Have to count tokens in a line--assumes 
            openDataFile();   // no extra spaces
            linetok = new StringTokenizer(datastream.readLine(), dataDelim);
            recordCount = linetok.countTokens();
            closeDataFile();
        }
        else recordCount = lineOffsets.size();
        return recordCount;
    }

    
    private void openDataFile() throws IOException {
        datastream = new DataInputStream(
            new BufferedInputStream(dataFile.openStream()));
    }

    private void closeDataFile() { datastream = null; }

    private Class pickType(String declaredType) {
        // We may want to make this more fancy in the future, taking into
        // account range, etc.
        Class retClass = null;
        try {
            if (declaredType.equalsIgnoreCase("INT")) 
                retClass = Class.forName("java.lang.Integer");
            else if (declaredType.equalsIgnoreCase("LONG")) 
                retClass = Class.forName("java.lang.Long");
            else if (declaredType.equalsIgnoreCase("STRING")) 
                retClass = Class.forName("java.lang.String");
            else if (declaredType.equalsIgnoreCase("DEC")) 
                retClass = Class.forName("java.lang.Float");
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        return retClass;
    }

    private static void skipToOffset(DataInputStream dis, long off) 
        throws IOException {
        for (long i = 0 ; i < off; i += dis.skip(off - i));
    }
        
    private static Object classToType(String s, Class storedType) 
        throws NumberFormatException {
        Object returnObj = null;
        char hashVal = storedType.getName().charAt(10);
        switch (hashVal) {
          case 'I': returnObj = new Integer(s); break;
          case 'L': returnObj = new Long(s); break;
          case 'F': returnObj = new Float(s); break;
          case 'D': returnObj = new Double(s); break;
          case 'S': returnObj = s; break;
        }
        return returnObj;
    }

    private int varIndex(String varName) throws DatabaseException {
        // helper method
        Object indexObj = varIndices.get(varName);
        if (indexObj == null) 
            throw new DatabaseException("No such variable.");
        return ((Integer)indexObj).intValue();
    }

    private Variable getVariable(int varIndex) {
        // helper method
        return ((Variable)(variables.elementAt(varIndex)));
    }

    /**
     * Helper function for parseCodebookFile.  
     * TODO: Write a real parser with lex & yacc
     */
    private static Tag getNextTag(LineNumberInputStream is) 
        throws IOException, DatabaseException {
        Tag returnTag = null;
        StreamTokenizer st = new StreamTokenizer(is);
        st.ordinaryChar('/'); // For some reason, this is treated weird otherwise

        while (st.nextToken() != '<')      // Assumes EOF won't happen inside a tag
            if (st.ttype == st.TT_EOF) return new Tag("EOF");

        if (st.nextToken() == '/') {    // It's an ending tag
            if (st.nextToken() != st.TT_WORD)
                throw new CodebookSyntaxException(is.getLineNumber());
            returnTag = new Tag("END_" + st.sval.toUpperCase());
            if (st.nextToken() != '>')
                throw new CodebookSyntaxException(is.getLineNumber());
        }
        else {      // It's a beginning tag
            if (st.ttype != st.TT_WORD)
                throw new CodebookSyntaxException(is.getLineNumber());
            returnTag = new Tag(st.sval.toUpperCase());     // Get name
            while (st.nextToken() != '>')  { // Get attlist for tag
                if (st.ttype != st.TT_WORD)
                    throw new CodebookSyntaxException(is.getLineNumber());
                String attname = st.sval.toUpperCase();
                if (st.nextToken() != '=') 
                    throw new CodebookSyntaxException(is.getLineNumber());
                if (st.nextToken() == st.TT_NUMBER) 
                    returnTag.attlist.put(attname, "" + st.nval);
                else if (st.sval == null) 
                    throw new CodebookSyntaxException(is.getLineNumber());
                else returnTag.attlist.put(attname, st.sval);
            }
        }
        return returnTag;
    }
}


/**
 * Spits out token types as it tokenizes.  Used for debugging.
 */
class DebugTokenizer extends StreamTokenizer {
    public DebugTokenizer(InputStream is) { super(is); }
    public int nextToken() throws IOException {         // Overrides parent version
        System.out.println("ttype is " + super.nextToken());
        return ttype;
    }
}


